from __future__ import annotations

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Tuple

import getml
from getml.pipeline import Pipeline
from mlflow.client import MlflowClient

from getml_mlflow.constants import DEFAULT_GETML_PROJECTS_PATH


def log_pipeline_as_artifact(
    mlflow_client: MlflowClient,
    run_id: str,
    pipeline: Pipeline,
    *,
    project_name: Optional[str] = None,
    projects_path: Path = DEFAULT_GETML_PROJECTS_PATH,
) -> str:
    if project_name is None:
        project_name = getml.project.name

    pipeline._save()

    artifact_path: str = f"pipeline/{project_name}"

    for item in (projects_path / project_name).iterdir():
        if item.is_dir():
            if item.stem == "data":
                continue
            if item.stem == "pipelines":
                mlflow_client.log_artifact(
                    run_id,
                    item / pipeline.id,
                    f"{artifact_path}/pipelines",
                )
                continue

        mlflow_client.log_artifact(run_id, item, artifact_path)

    return pipeline.id


def download_artifact_pipeline(
    mlflow_client: MlflowClient,
    run_id: str,
    pipeline_id: str,
    *,
    original_project_name: Optional[str] = None,
    projects_path: Path = DEFAULT_GETML_PROJECTS_PATH,
) -> Tuple[str, str]:
    """
    Downloads a getML pipeline artifact from an MLflow run and saves it as a new project.

    This function downloads a pipeline artifact from MLflow and creates a new project with
    a name derived from the original project name and the pipeline ID: 
    "original_project_name-pipeline_id".

    If the project already exists (e.g., when calling this function multiple times with 
    the same parameters), the existing project will be overwritten with the downloaded 
    artifacts.

    ??? warning "Experimental feature"
        This feature is experimental and may change in future releases.

    Args:
        mlflow_client: An MLflow client instance to interact with MLflow.

        run_id: The ID of the MLflow run containing the pipeline artifacts.

        pipeline_id: The ID of the pipeline to be downloaded.

        original_project_name: The name of the original getML project the pipeline was 
            saved from. If None, uses the current project name.

        projects_path: Path where getML projects are stored. Defaults to 
            `$HOME/.getML/projects`.

    Returns:
        A tuple containing:
            - The name of the newly created getML project
            - The ID of the downloaded pipeline

    Example:
        ```python
        # Initialize MLflow client
        client = MlflowClient(tracking_uri="http://localhost:5000")

        run_id = "abcdef1234567890"
        pipeline_id = "l2TCiD"

        # Download pipeline artifact from a specific run. This creates a new project 
        # named "interstate94-l2TCiD" with the pipeline
        new_project, pipeline_id = getml_mlflow.marshalling.pipeline.download_artifact_pipeline(
            client, run_id, pipeline_id, original_project_name="interstate94"
            )

        # You can now switch to the new project and load the pipeline
        getml.project.set_project(new_project)
        pipeline = getml.pipeline.load(pipeline_id)
        ```

    """
    
    if original_project_name is None:
        original_project_name = getml.project.name

    with TemporaryDirectory() as temp_dir:
        mlflow_client.download_artifacts(
            run_id, f"pipeline/{original_project_name}", temp_dir
        )
        new_project_name: str = f"{original_project_name}-{pipeline_id}"
        project_path: Path = projects_path / new_project_name
        temp_project_path: Path = Path(temp_dir) / "pipeline" / original_project_name
        temp_project_path.rename(project_path)

    return (new_project_name, pipeline_id)


def switch_to_artifact_pipeline(
    mlflow_client: MlflowClient,
    run_id: str,
    pipeline_id: str,
    *,
    original_project_name: Optional[str] = None,
    projects_path: Path = DEFAULT_GETML_PROJECTS_PATH,
) -> Pipeline:
    """
    Downloads an artifact pipeline from MLflow, switches to the newly created project,
    and loads the pipeline.

    This function simplifies the workflow of retrieving a pipeline stored as an MLflow 
    artifact. It downloads the pipeline into a new getML project (named as 
    "original_project_name-pipeline_id"), automatically switches to that project, and 
    loads the pipeline for immediate use.

    ??? warning "Experimental feature"
        This function is experimental and may change in future releases.

    Args:
        mlflow_client: The MLflow client instance to use for retrieving the artifact.

        run_id: The ID of the MLflow run containing the pipeline artifact.

        pipeline_id: The ID of the pipeline to download.

        original_project_name: The name of the original project. If None, the current 
            project name is used. Defaults to None.

        projects_path: Path to the getML projects directory. Defaults to 
            `$HOME/.getML/projects`.

    Returns:
        Pipeline: The loaded pipeline object in the newly created project.

    Examples:
        ```python
        import mlflow
        from mlflow.tracking import MlflowClient
        import getml_mlflow
        import getml
        
        # Connect to MLflow
        client = MlflowClient("http://localhost:5000")
        
        # Download pipeline from run and switch to new project
        pipeline = getml_mlflow.marshalling.pipeline.switch_to_artifact_pipeline(
                   client,
                   "2960ee40202744daa64aa83d180f0b2f",
                   "uPe3hR"
                )
        
        # Pipeline is ready to use
        predictions = pipeline.predict(container.test)
        ```

    """
    if original_project_name is None:
        original_project_name = getml.project.name

    project_name, pipeline_id = download_artifact_pipeline(
        mlflow_client=mlflow_client,
        run_id=run_id,
        pipeline_id=pipeline_id,
        original_project_name=original_project_name,
        projects_path=projects_path,
    )
    getml.project.switch(project_name)
    return getml.pipeline.load(pipeline_id)



from __future__ import annotations

from enum import Enum
from functools import cached_property
from tempfile import TemporaryDirectory
from typing import Dict, List, Sequence, Union

from getml.data import Subset
from mlflow import MlflowClient
from mlflow.entities import DatasetInput, InputTag, Run
from mlflow.utils.mlflow_tags import MLFLOW_DATASET_CONTEXT

from getml_mlflow.data import dataframelike
from getml_mlflow.data.dataframelike import DataFrameLike, DataFrameLikeT
from getml_mlflow.data.getml_dataset import GetMLDataset
from getml_mlflow.loggingconfiguration import DataContainerLoggingConfiguration


class DataContainerLoggerTarget(str, Enum):
    ARTIFACT = "artifact"
    INPUT = "input"


class DataContainerLogger:
    @classmethod
    def as_artifact(
        cls,
        mlflow_client: MlflowClient,
        run_id: str,
        *,
        logging_configuration: DataContainerLoggingConfiguration = DataContainerLoggingConfiguration(),
    ) -> DataContainerLogger:
        return cls(
            mlflow_client,
            run_id,
            DataContainerLoggerTarget.ARTIFACT,
            logging_configuration=logging_configuration,
        )

    @classmethod
    def as_input(
        cls,
        mlflow_client: MlflowClient,
        run_id: str,
        *,
        logging_configuration: DataContainerLoggingConfiguration = DataContainerLoggingConfiguration(),
    ) -> DataContainerLogger:
        return cls(
            mlflow_client,
            run_id,
            DataContainerLoggerTarget.INPUT,
            logging_configuration=logging_configuration,
        )

    def __init__(
        self,
        mlflow_client: MlflowClient,
        run_id: str,
        target: DataContainerLoggerTarget,
        *,
        logging_configuration: DataContainerLoggingConfiguration = DataContainerLoggingConfiguration(),
    ) -> None:
        self._mlflow_client: MlflowClient = mlflow_client
        self._run_id: str = run_id
        self._run: Run = self._mlflow_client.get_run(run_id)
        self._target: DataContainerLoggerTarget = target
        self._logging_configuration: DataContainerLoggingConfiguration = (
            logging_configuration
        )

    @cached_property
    def _separator(self) -> str:
        if self._target == DataContainerLoggerTarget.ARTIFACT:
            return "/"
        elif self._target == DataContainerLoggerTarget.INPUT:
            return "."
        else:
            raise ValueError(f"Unknown target: {self._target}")

    def _log_dataframe_like(
        self, dataframe_like: DataFrameLike, context: List[str]
    ) -> None:
        if self._target == DataContainerLoggerTarget.ARTIFACT:
            return self._log_dataframe_like_as_artifact(dataframe_like, context)
        elif self._target == DataContainerLoggerTarget.INPUT:
            return self._log_dataframe_like_as_input(dataframe_like, context)
        else:
            raise ValueError(f"Unknown target: {self._target}")

    def log_data_containers(
        self,
        data_containers: Union[
            Sequence[DataFrameLike],
            Dict[str, DataFrameLike],
        ],
        prefix: str,
    ) -> None:
        if isinstance(data_containers, dict):
            for name, data_container in data_containers.items():
                self.log_data_container(data_container, [prefix, name])
        else:
            for id, data_container in enumerate(data_containers):
                self.log_data_container(data_container, [prefix, str(id)])

    def log_data_container(
        self,
        data_container: Union[DataFrameLike, Subset],
        context: Union[str, List[str]],
    ) -> None:
        if isinstance(context, str):
            context = [context]
        if isinstance(data_container, DataFrameLikeT):
            self._log_dataframe_like(data_container, context)
        elif isinstance(data_container, Subset):
            # Subset is a convenience case, usually used to transfer
            # population and peripheral tables together as Population
            # and then internally split into Population and Peripheral
            # Therefore we remove the last context element, usually "Population"
            self._log_subset(data_container, context[:-1])

    def _log_dataframe_like_as_input(
        self, dataframe_like: DataFrameLike, context: List[str]
    ) -> None:
        if not self._logging_configuration.log_information:
            return

        dataset: GetMLDataset = GetMLDataset(dataframe_like)

        dataset_context: str = self._separator.join(context)
        dataset_input: DatasetInput = DatasetInput(
            dataset=dataset._to_mlflow_entity(),
            tags=[InputTag(key=MLFLOW_DATASET_CONTEXT, value=dataset_context)],
        )
        self._mlflow_client.log_inputs(
            run_id=self._run_id,
            datasets=[dataset_input],
        )
        self._mlflow_client.log_dict(
            self._run_id,
            dataset.as_dict(),
            f"input/dataset/{dataset_context}.{dataset.name}.json",
        )

    def _log_dataframe_like_as_artifact(
        self, dataframe_like: DataFrameLike, context: List[str]
    ) -> None:
        artifact_path: str = self._separator.join(context)
        if self._logging_configuration.log_as_artifact:
            with TemporaryDirectory() as temp_dir:
                filename: str = dataframelike.get_name(dataframe_like) + ".parquet"
                local_path: str = f"{temp_dir}/{filename}"
                dataframe_like.to_parquet(local_path)
                self._mlflow_client.log_artifact(
                    run_id=self._run_id,
                    local_path=local_path,
                    artifact_path=artifact_path,
                )
        if self._logging_configuration.log_information:
            dataset: GetMLDataset = GetMLDataset(dataframe_like)
            self._mlflow_client.log_dict(
                self._run_id, dataset.as_dict(), f"{artifact_path}/{dataset.name}.json"
            )

    def _log_subset(self, subset: Subset, context: List[str]) -> None:
        self._log_dataframe_like(subset.population, context + ["Population"])

        for name, table in subset.peripheral.items():
            self._log_dataframe_like(table, context + ["Peripheral", name])



from __future__ import annotations

from typing import Union

from getml.data import DataFrame, View

DataFrameLike = Union[DataFrame, View]
DataFrameLikeT = (DataFrame, View)


def get_name(dataframe_like: DataFrameLike) -> str:
    if isinstance(dataframe_like, DataFrame):
        return str(dataframe_like.name)
    else:
        return f"{get_dataframe_name(dataframe_like)}.{dataframe_like.name}"


def get_dataframe_name(dataframe_like: DataFrameLike) -> str:
    return str(get_base(dataframe_like).name)


def get_base(dataframe_like: DataFrameLike) -> DataFrame:
    if isinstance(dataframe_like, DataFrame):
        return dataframe_like
    return get_base(dataframe_like.base)



from __future__ import annotations

import hashlib
import json
from functools import cached_property
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Union

import getml
from getml.data import DataFrame, Roles
from getml.data.roles.types import Role
from mlflow.data.dataset import Dataset
from mlflow.data.dataset_source import DatasetSource
from mlflow.types.schema import ColSpec, DataType, Schema
from typing_extensions import override

from getml_mlflow.constants import DEFAULT_GETML_HOST, DEFAULT_GETML_PORT
from getml_mlflow.data import dataframelike
from getml_mlflow.data.dataframelike import DataFrameLike, DataFrameLikeT


class GetMLDatasetSource(DatasetSource):
    @classmethod
    def from_parquet(
        cls,
        path: str,
        roles: Union[Dict[Union[Role, str], Iterable[str]], Roles, None] = None,
        ignore: bool = False,
        colnames: Iterable[str] = (),
    ) -> GetMLDatasetSource:
        dataframe: DataFrame = DataFrame.from_parquet(
            path, Path(path).stem, roles, ignore, False, colnames
        )
        return cls(dataframe)

    @classmethod
    def from_getml(
        cls,
        dataframe_name: str,
        roles: Union[Dict[Union[Role, str], Iterable[str]], Roles, None] = None,
    ) -> GetMLDatasetSource:
        dataframe: DataFrame = DataFrame(dataframe_name, roles).load()
        return cls(dataframe)

    @classmethod
    def from_dataframe_like(cls, dataframe_like: DataFrameLike) -> GetMLDatasetSource:
        dataframe: DataFrame = dataframelike.get_base(dataframe_like).save().load()
        return cls(dataframe)

    def __init__(self, dataframe: DataFrame) -> None:
        self._dataframe: DataFrame = dataframe
        super().__init__()

    @override
    @staticmethod
    def _get_source_type() -> str:
        return "http"

    @override
    def load(self) -> Any:
        return self._dataframe.save().load()

    @override
    @staticmethod
    def _can_resolve(raw_source: Any) -> bool:
        return isinstance(raw_source, (str, DataFrameLikeT))

    @override
    @classmethod
    def _resolve(cls, raw_source: Any) -> GetMLDatasetSource:
        if isinstance(raw_source, DataFrameLikeT):
            return cls.from_dataframe_like(raw_source)
        if isinstance(raw_source, str):
            if raw_source.endswith(".parquet"):
                return cls.from_parquet(raw_source)

            return cls.from_getml(raw_source)
        raise NotImplementedError(f"Cannot resolve source {raw_source}")

    @override
    def to_dict(self) -> dict:
        project_name: str = getml.project.name
        dataframe_name: str = dataframelike.get_dataframe_name(self._dataframe)

        return {
            # TODO: Extract URL to a default URL and make it configurable
            "url": f"http://{DEFAULT_GETML_HOST}:{DEFAULT_GETML_PORT}/#/getdataframe/{project_name}/{dataframe_name}/",
            "dataframe_name": dataframe_name,
            "project_name": project_name,
            "roles": self._dataframe.roles.to_dict(),
        }

    @override
    @classmethod
    def from_dict(cls, source_dict: dict) -> GetMLDatasetSource:
        return cls.from_getml(source_dict["dataframe_name"], source_dict["roles"])


class GetMLDataset(Dataset):
    GETML_ROLE_TO_MLFLOW_TYPE: Dict[str, DataType] = {
        "categorical": DataType.string,
        "join_key": DataType.string,
        "numerical": DataType.double,
        "target": DataType.double,
        "text": DataType.string,
        "time_stamp": DataType.double,
        "unused_float": DataType.double,
        "unused_string": DataType.string,
    }

    GETML_ROLE_TO_EMOJI: Dict[str, str] = {
        "categorical": "🗃",
        "join_key": "🔗",
        "numerical": "🔢",
        "target": "🎯",
        "text": "📝",
        "time_stamp": "⏰",
        "unused_float": "🧮",
        "unused_string": "🧵",
    }

    def __init__(
        self,
        dataframe_like: DataFrameLike,
        source: Optional[DatasetSource] = None,
        name: Optional[str] = None,
        digest: Optional[str] = None,
    ) -> None:
        self._dataframe_like = dataframe_like
        self._name: str = (
            name if name is not None else dataframelike.get_name(dataframe_like)
        )
        resolved_source: DatasetSource = (
            source if source is not None else self._resolve_source(dataframe_like)
        )
        super().__init__(resolved_source, self._name, digest)

    @override
    def _compute_digest(self) -> str:
        return hashlib.md5(json.dumps(self._as_base_dict()).encode()).hexdigest()[:8]

    @override
    def to_dict(self) -> Dict[str, str]:
        result: Dict[str, str] = super().to_dict()
        result.update(
            {
                "profile": json.dumps(self.profile),
                "schema": (
                    json.dumps({"mlflow_colspec": self.schema.to_dict()})
                    if self.schema
                    else ""
                ),
            }
        )
        return result

    def _as_base_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "source": self.source.to_dict(),
            "source_type": self.source._get_source_type(),
            "schema": self.schema.to_dict() if self.schema else None,
            "profile": self.profile,
            "roles": self._roles,
        }

    def as_dict(self) -> Dict[str, Any]:
        result: Dict[str, Any] = self._as_base_dict()
        result.update(
            {
                "digest": self.digest,
            }
        )
        return result

    @property
    @override
    def profile(self) -> Optional[Any]:
        ncols: int = self._dataframe_like.ncols()
        nrows: Union[int, str] = self._dataframe_like.nrows()
        return {
            "num_rows": nrows,
            "num_cols": ncols,
            "num_elements": nrows if isinstance(nrows, str) else nrows * ncols,
            "base_name": dataframelike.get_dataframe_name(self._dataframe_like),
        }

    @property
    @override
    def schema(self) -> Optional[Any]:
        return Schema(
            [
                self._to_colspec(name, role)
                for (name, role) in self._dataframe_like.roles.to_mapping().items()
            ]
        )

    def _to_colspec(self, name: str, role: str) -> ColSpec:
        return ColSpec(
            type=self.GETML_ROLE_TO_MLFLOW_TYPE[role],
            name=f"{self.GETML_ROLE_TO_EMOJI[role]} {name}",
            required=not role.startswith("unused_"),
        )

    def _resolve_source(self, dataframe_like: DataFrameLike) -> DatasetSource:
        return GetMLDatasetSource.from_dataframe_like(dataframe_like)

    @cached_property
    def _roles(self) -> Dict[str, str]:
        return {
            key: str(value)
            for (key, value) in self._dataframe_like.roles.to_mapping().items()
        }



from __future__ import annotations

import json
import warnings
from dataclasses import is_dataclass
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Protocol,
    Type,
    Union,
    runtime_checkable,
)

from getml.pipeline import Pipeline, Scores
from mlflow import MlflowClient
from mlflow.entities import Metric, Param, RunTag
from mlflow.utils.time import get_current_time_millis
from pydantic import TypeAdapter

from getml_mlflow.constants import DEFAULT_GETML_PROJECTS_PATH
from getml_mlflow.logging.logger import log_exit_exception
from getml_mlflow.loggingconfiguration import PipelineLoggingConfiguration
from getml_mlflow.marshalling.pipeline import log_pipeline_as_artifact
from getml_mlflow.util.callableenum import CallableEnumFactory


@runtime_checkable
class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, Any]]


PipelineParameter = CallableEnumFactory[Callable[[Pipeline], Any]].build(
    "PipelineParameter",
    dict(
        PREPROCESSORS=lambda pipeline: pipeline.preprocessors,
        FEATURE_LEARNERS=lambda pipeline: pipeline.feature_learners,
        FEATURE_SELECTORS=lambda pipeline: pipeline.feature_selectors,
        PREDICTORS=lambda pipeline: pipeline.predictors,
        LOSS_FUNCTION=lambda pipeline: pipeline.loss_function,
        INCLUDE_CATEGORICAL=lambda pipeline: pipeline.include_categorical,
        SHARE_SELECTED_FEATURES=lambda pipeline: pipeline.share_selected_features,
    ),
)


class PipelineLogger:
    def __init__(
        self,
        mlflow_client: MlflowClient,
        run_id: str,
        pipeline: Pipeline,
        *,
        logging_configuration: PipelineLoggingConfiguration = PipelineLoggingConfiguration(),
        getml_project_path: Path = DEFAULT_GETML_PROJECTS_PATH,
    ) -> None:
        self._mlflow_client: MlflowClient = mlflow_client
        self._run_id: str = run_id
        self._pipeline: Pipeline = pipeline
        self._logging_configuration: PipelineLoggingConfiguration = (
            logging_configuration
        )
        self._getml_project_path: Path = getml_project_path

    def log_constructor_arguments(self) -> None:
        self.log_parameters()
        self.log_tags()
        self.log_data_model()

    def log_generated_information(self) -> None:
        self.log_scores()
        self.log_features()
        self.log_columns()
        self.log_targets()
        self.log_pipeline_as_artifact()

    def __enter__(self) -> PipelineLogger:
        self.log_constructor_arguments()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        if exc_type is not None and exc_value is not None:
            log_exit_exception(
                self._mlflow_client, self._run_id, exc_type, exc_value, exc_traceback
            )
        else:
            self.log_generated_information()

    def log_parameters(self) -> None:
        if not self._logging_configuration.log_parameters:
            return

        parameters: List[Param] = []

        for pipeline_parameter in PipelineParameter:
            parameter: Any = pipeline_parameter.value(self._pipeline)
            parameter_name: str = pipeline_parameter.name.lower()
            if isinstance(parameter, list):
                assert all(map(is_dataclass, parameter))
                for id, item in enumerate(parameter):
                    parameters.extend(
                        self._serialize_dataclass(f"{parameter_name}.{id}", item)
                    )
            else:
                parameters.append(Param(parameter_name, str(parameter)))

        self._mlflow_client.log_batch(run_id=self._run_id, params=parameters)

    def _serialize_dataclass(
        self, name: str, parameter: DataclassInstance
    ) -> List[Param]:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                module="pydantic",
            )
            parameter_as_dict: Dict[str, Any] = TypeAdapter(
                type(parameter)
            ).dump_python(parameter)
        current_name: str = f"{name}.{type(parameter).__name__}"
        return self._to_param_list(
            current_name,
            self._flatten_parameters(parameter_as_dict),
        )

    def _flatten_parameters(self, parameters: Dict[str, Any]) -> Dict[str, Any]:
        flattened_parameters: Dict[str, Any] = {}
        for key, value in parameters.items():
            if isinstance(value, dict):
                for sub_key, sub_value in self._flatten_parameters(value).items():
                    flattened_parameters[f"{key}.{sub_key}"] = sub_value
            else:
                flattened_parameters[key] = value
        return flattened_parameters

    def _to_param_list(self, name: str, parameters: Dict[str, Any]) -> List[Param]:
        return [
            Param(f"{name}.{key}", self._serialize_field_value(value))
            for key, value in parameters.items()
        ]

    def _serialize_field_value(self, field_value: Any) -> str:
        if isinstance(field_value, (frozenset, set)):
            return json.dumps(sorted(field_value))
        if not isinstance(field_value, str):
            return json.dumps(field_value)
        return field_value

    def log_tags(self) -> None:
        if not self._logging_configuration.log_tags:
            return

        tags: List[RunTag] = [RunTag("id", self._pipeline.id)]
        for tag in map(str, self._pipeline.tags):
            if ":" in tag:
                key, value = tag.split(":")
                tags.append(RunTag(key.strip(), value.strip()))
            else:
                tags.append(RunTag(tag, tag))

        self._mlflow_client.log_batch(run_id=self._run_id, tags=tags)

    def log_scores(self) -> None:
        if not self._logging_configuration.log_scores:
            return

        metrics: List[Metric] = []

        scores: Scores = self._pipeline.scores

        if self._pipeline.is_classification:
            metrics.extend(self._serialize_metric("auc", scores.auc))
            metrics.extend(self._serialize_metric("accuracy", scores.accuracy))
            metrics.extend(
                self._serialize_metric("cross_entropy", scores.cross_entropy)
            )

        if self._pipeline.is_regression:
            metrics.extend(self._serialize_metric("mae", scores.mae))
            metrics.extend(self._serialize_metric("rmse", scores.rmse))
            metrics.extend(self._serialize_metric("rsquared", scores.rsquared))

        self._mlflow_client.log_batch(run_id=self._run_id, metrics=metrics)

    # TODO: Find better way to log and display features
    def log_features(self) -> None:
        if not self._logging_configuration.log_features:
            return

        data: Dict[str, List[Union[int, float, str]]] = {
            "feature-names": [feature.name for feature in self._pipeline.features],
            "feature-importances": [
                feature.importance for feature in self._pipeline.features
            ],
            "feature-correlations": [
                feature.correlation for feature in self._pipeline.features
            ],
        }
        self._mlflow_client.log_table(
            self._run_id, data, f"output/pipeline.{self._pipeline.id}.features.json"
        )

    # TODO: Find better way to log and display columns
    def log_columns(self) -> None:
        if not self._logging_configuration.log_columns:
            return

        data: Dict[str, List[Union[int, float, str]]] = {
            "column-names": [column.name for column in self._pipeline.columns],
            "column-table": [column.table for column in self._pipeline.columns],
            "column-importances": [
                column.importance for column in self._pipeline.columns
            ],
        }
        self._mlflow_client.log_table(
            self._run_id, data, f"output/pipeline.{self._pipeline.id}.columns.json"
        )

    def log_targets(self) -> None:
        if not self._logging_configuration.log_targets:
            return

        self._mlflow_client.log_param(self._run_id, "targets", self._pipeline.targets)

    def _serialize_metric(self, name: str, values: float | List[float]) -> List[Metric]:
        timestamp: int = get_current_time_millis()
        if isinstance(values, list):
            return [
                Metric(
                    key=f"{name}.{id}",
                    value=value,
                    timestamp=timestamp,
                    step=0,
                )
                for id, value in enumerate(values)
            ]
        return [
            Metric(
                key=name,
                value=values,
                timestamp=timestamp,
                step=0,
            )
        ]

    def log_pipeline_as_artifact(self) -> None:
        if not self._logging_configuration.log_as_artifact:
            return

        log_pipeline_as_artifact(
            mlflow_client=self._mlflow_client,
            run_id=self._run_id,
            pipeline=self._pipeline,
            projects_path=self._getml_project_path,
        )

    def log_data_model(self) -> None:
        if not self._logging_configuration.log_data_model:
            return

        self._mlflow_client.log_text(
            run_id=self._run_id,
            text=self._pipeline.data_model._repr_html_(),
            artifact_file="input/data_model.html",
        )



from __future__ import annotations

import getpass
import socket
from types import TracebackType
from typing import Callable, Dict, Optional, Sequence, Type, Union

import getml
import mlflow
import mlflow.entities
import numpy
from getml.data import DataFrame, Subset
from getml.pipeline import Pipeline, Scores
from mlflow import MlflowClient
from mlflow.entities import RunStatus, RunTag
from mlflow.utils.mlflow_tags import MLFLOW_PARENT_RUN_ID, MLFLOW_USER
from numpy.typing import NDArray

from getml_mlflow.constants import DEFAULT_GETML_PROJECTS_PATH
from getml_mlflow.data.dataframelike import DataFrameLike
from getml_mlflow.logging.function import FunctionLogger
from getml_mlflow.logging.logger import log_exit_exception
from getml_mlflow.logging.pipeline import PipelineLogger
from getml_mlflow.logging.systemmetrics import SystemMetricsLogger
from getml_mlflow.loggingconfiguration import LoggingConfiguration


class Run:
    def __init__(
        self,
        mlflow_client: MlflowClient,
        pipeline: Pipeline,
        name: str,
        *,
        create_runs: bool = True,
        extra_tags: Optional[Dict[str, str]] = None,
    ) -> None:
        self._pipeline: Pipeline = pipeline
        self._mlflow_client: MlflowClient = mlflow_client
        self._run: Optional[mlflow.entities.Run] = None
        self._name: str = name
        self._create_runs: bool = create_runs
        self._extra_tags: Dict[str, str] = extra_tags or {}

    def __enter__(self) -> Run:
        if not self._create_runs:
            self._run = mlflow.active_run()
            if self._run is None:
                raise RuntimeError("No active MLflow run found.")
            self._log_extra_tags()
            return self

        create_run_args: dict = {
            "experiment_id": self._experiment_id(),
            "run_name": self._name,
            "tags": {
                "id": self._pipeline.id,
                MLFLOW_USER: f"{getpass.getuser()}@{socket.gethostname()}",
            },
        }
        if parent_run_id := self._parent_run_id():
            create_run_args["tags"].update(
                {
                    MLFLOW_PARENT_RUN_ID: parent_run_id,
                }
            )
        self._run = self._mlflow_client.create_run(**create_run_args)
        self._log_extra_tags()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        if exc_type is not None and exc_value is not None:
            log_exit_exception(
                self._mlflow_client, self.id, exc_type, exc_value, exc_traceback
            )
            if self._create_runs:
                self._mlflow_client.set_terminated(
                    self.id, status=RunStatus.to_string(RunStatus.FAILED)
                )
        else:
            if self._create_runs:
                self._mlflow_client.set_terminated(
                    self.id, status=RunStatus.to_string(RunStatus.FINISHED)
                )
        self._run = None

    def _experiment_id(self) -> str:
        if run_info := getattr(self._pipeline, "_mlflow_run_info"):
            return run_info.experiment_id
        else:
            project_name: str = getml.project.name
            if experiment := self._mlflow_client.get_experiment_by_name(project_name):
                return experiment.experiment_id

            raise LookupError(f"MLflow Experiment '{project_name}' not found")

    def _parent_run_id(self) -> Optional[str]:
        if run_info := getattr(self._pipeline, "_mlflow_run_info"):
            return run_info.run_id

        return None

    @property
    def id(self) -> str:
        assert self._run, "RUN is missing. Make sure to be inside a context manager."
        return self._run.info.run_id

    @property
    def info(self) -> mlflow.entities.RunInfo:
        assert self._run, "RUN is missing. Make sure to be inside a context manager."
        return self._run.info

    def _log_extra_tags(self) -> None:
        if self._extra_tags:
            self._mlflow_client.log_batch(
                self.id,
                tags=[RunTag(*item) for item in self._extra_tags.items()],
            )

    @property
    def run(self) -> mlflow.entities.Run:
        assert self._run, "RUN is missing. Make sure to be inside a context manager."
        return self._run


def init(original: Callable, pipeline: Pipeline, *args, **kwargs) -> None:
    init_method: Callable = original
    if not hasattr(pipeline, "_mlflow_run_info"):
        setattr(pipeline, "_mlflow_run_info", None)

    init_method(pipeline, *args, **kwargs)


def load(
    original: Callable,
    name: str,
    *,
    logging_configuration: LoggingConfiguration = LoggingConfiguration(),
) -> Pipeline:
    mlflow_client: MlflowClient = logging_configuration.general.mlflow_client
    load_function: Callable = original

    load_output: Pipeline = load_function(name)

    with Run(
        mlflow_client=mlflow_client,
        pipeline=load_output,
        name=pipeline_name(load_output),
        create_runs=logging_configuration.general.create_runs,
        extra_tags=logging_configuration.general.extra_tags,
    ) as run:
        setattr(load_output, "_mlflow_run_info", run.info)

    return load_output


def pipeline_name(pipeline: Pipeline) -> str:
    return f"Pipeline-{pipeline.id.replace(' ', '_')}"


def fit(
    original: Callable,
    pipeline: Pipeline,
    population_table: Union[DataFrameLike, Subset],
    peripheral_tables: Optional[
        Union[
            Sequence[DataFrameLike],
            dict[str, DataFrameLike],
        ]
    ] = None,
    validation_table: Optional[Union[DataFrameLike, Subset]] = None,
    check: bool = True,
    *,
    logging_configuration: LoggingConfiguration = LoggingConfiguration(),
) -> Pipeline:
    mlflow_client: MlflowClient = logging_configuration.general.mlflow_client
    fit_method: Callable = original

    ###############
    # Log pipeline
    with Run(
        mlflow_client=mlflow_client,
        pipeline=pipeline,
        name=pipeline_name(pipeline),
        create_runs=logging_configuration.general.create_runs,
        extra_tags=logging_configuration.general.extra_tags,
    ) as run:
        setattr(pipeline, "_mlflow_run_info", run.info)
        pipeline_run_id: str = run.info.run_id
        pipeline_logger: PipelineLogger = PipelineLogger(
            mlflow_client,
            run.id,
            pipeline,
            logging_configuration=logging_configuration.pipeline,
            getml_project_path=logging_configuration.general.getml_project_path
            or DEFAULT_GETML_PROJECTS_PATH,
        )
        pipeline_logger.log_constructor_arguments()

    ###############
    # Log fit call
    with Run(
        mlflow_client=mlflow_client,
        pipeline=pipeline,
        name="fit",
        create_runs=logging_configuration.general.create_runs,
        extra_tags=logging_configuration.general.extra_tags,
    ) as fit_run:
        with PipelineLogger(
            mlflow_client,
            fit_run.id,
            pipeline,
            logging_configuration=logging_configuration.pipeline,
            getml_project_path=logging_configuration.general.getml_project_path
            or DEFAULT_GETML_PROJECTS_PATH,
        ):
            with SystemMetricsLogger(
                mlflow_client,
                fit_run.id,
                log_system_metrics=logging_configuration.general.log_system_metrics,
            ):
                fit_output: Pipeline = FunctionLogger(
                    mlflow_client,
                    fit_run.run,
                    pipeline,
                    function_logging_configuration=logging_configuration.function,
                    data_container_logging_configuration=logging_configuration.data_container,
                ).log(fit_method)(
                    pipeline,
                    population_table,
                    peripheral_tables,
                    validation_table,
                    check,
                )
            mlflow_client.set_tag(fit_run.id, "id", pipeline.id)

    ##################
    # Update pipeline
    mlflow_client.set_tag(pipeline_run_id, "id", pipeline.id)
    if logging_configuration.general.create_runs:
        mlflow_client.update_run(
            run_id=pipeline_run_id,
            name=pipeline_name(pipeline),
        )

    return fit_output


def score(
    original: Callable,
    pipeline: Pipeline,
    population_table: Union[DataFrameLike, Subset],
    peripheral_tables: Optional[
        Union[
            Sequence[DataFrameLike],
            Dict[str, DataFrameLike],
        ]
    ] = None,
    *,
    logging_configuration: LoggingConfiguration = LoggingConfiguration(),
) -> Scores:
    score_method: Callable = original
    mlflow_client: MlflowClient = logging_configuration.general.mlflow_client

    with Run(
        mlflow_client=mlflow_client,
        pipeline=pipeline,
        name="score",
        create_runs=logging_configuration.general.create_runs,
        extra_tags=logging_configuration.general.extra_tags,
    ) as score_run:
        with PipelineLogger(
            mlflow_client,
            score_run.id,
            pipeline,
            logging_configuration=logging_configuration.pipeline,
            getml_project_path=logging_configuration.general.getml_project_path
            or DEFAULT_GETML_PROJECTS_PATH,
        ):
            score_output: Scores = FunctionLogger(
                mlflow_client,
                score_run.run,
                pipeline,
                function_logging_configuration=logging_configuration.function,
                data_container_logging_configuration=logging_configuration.data_container,
            ).log(score_method)(pipeline, population_table, peripheral_tables)

    return score_output


def predict(
    original: Callable,
    pipeline: Pipeline,
    population_table: Union[DataFrameLike, Subset],
    peripheral_tables: Optional[
        Union[
            Sequence[DataFrameLike],
            Dict[str, DataFrameLike],
        ]
    ] = None,
    table_name: str = "",
    *,
    logging_configuration: LoggingConfiguration = LoggingConfiguration(),
) -> Union[NDArray[numpy.float_], None]:
    mlflow_client: MlflowClient = logging_configuration.general.mlflow_client
    predict_method: Callable = original

    with Run(
        mlflow_client=mlflow_client,
        pipeline=pipeline,
        name="predict",
        create_runs=logging_configuration.general.create_runs,
        extra_tags=logging_configuration.general.extra_tags,
    ) as predict_run:
        with PipelineLogger(
            mlflow_client=mlflow_client,
            run_id=predict_run.id,
            pipeline=pipeline,
            logging_configuration=logging_configuration.pipeline,
            getml_project_path=logging_configuration.general.getml_project_path
            or DEFAULT_GETML_PROJECTS_PATH,
        ):
            predict_output: Union[NDArray[numpy.float_], None] = FunctionLogger(
                mlflow_client,
                predict_run.run,
                pipeline,
                function_logging_configuration=logging_configuration.function,
                data_container_logging_configuration=logging_configuration.data_container,
            ).log(
                predict_method
            )(pipeline, population_table, peripheral_tables, table_name)

    return predict_output


def transform(
    original: Callable,
    pipeline: Pipeline,
    population_table: Union[DataFrameLike, Subset],
    peripheral_tables: Optional[
        Union[
            Sequence[DataFrameLike],
            Dict[str, DataFrameLike],
        ]
    ] = None,
    df_name: str = "",
    table_name: str = "",
    *,
    logging_configuration: LoggingConfiguration = LoggingConfiguration(),
) -> Union[DataFrame, NDArray[numpy.float_], None]:
    mlflow_client: MlflowClient = logging_configuration.general.mlflow_client
    transform_method: Callable = original

    with Run(
        mlflow_client=mlflow_client,
        pipeline=pipeline,
        name="transform",
        create_runs=logging_configuration.general.create_runs,
        extra_tags=logging_configuration.general.extra_tags,
    ) as transform_run:
        with PipelineLogger(
            mlflow_client,
            transform_run.id,
            pipeline,
            logging_configuration=logging_configuration.pipeline,
            getml_project_path=logging_configuration.general.getml_project_path
            or DEFAULT_GETML_PROJECTS_PATH,
        ):
            transform_output: Union[DataFrame, NDArray[numpy.float_], None] = (
                FunctionLogger(
                    mlflow_client,
                    transform_run.run,
                    pipeline,
                    function_logging_configuration=logging_configuration.function,
                    data_container_logging_configuration=logging_configuration.data_container,
                ).log(transform_method)(
                    pipeline, population_table, peripheral_tables, df_name, table_name
                )
            )

    return transform_output



from __future__ import annotations

from typing import Callable

from mlflow import MlflowClient

from getml_mlflow.loggingconfiguration import LoggingConfiguration


def switch(
    original: Callable,
    name: str,
    *,
    logging_configuration: LoggingConfiguration = LoggingConfiguration(),
) -> None:
    mlflow_client: MlflowClient = logging_configuration.general.mlflow_client
    switch_function: Callable = original

    switch_function(name)

    if not mlflow_client.search_experiments(filter_string=f"name='{name}'"):
        mlflow_client.create_experiment(name=name)



from pathlib import Path

DEFAULT_GETML_PROJECTS_PATH: Path = Path.home() / ".getML" / "projects"
DEFAULT_GETML_HOST: str = "localhost"
DEFAULT_GETML_PORT: int = 1709
DEFAULT_MLFLOW_TRACKING_URI: str = "http://localhost:5000"



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Optional

from mlflow import MlflowClient


@dataclass
class DataContainerLoggingConfiguration:
    log_information: bool = True
    log_as_artifact: bool = True


@dataclass
class FunctionLoggingConfiguration:
    log_parameters: bool = True
    log_return: bool = True
    log_as_trace: bool = True


@dataclass
class PipelineLoggingConfiguration:
    log_parameters: bool = True
    log_tags: bool = True
    log_scores: bool = True
    log_features: bool = True
    log_columns: bool = True
    log_targets: bool = True
    log_data_model: bool = True
    log_as_artifact: bool = True


@dataclass
class GeneralLoggingConfiguration:
    mlflow_client: MlflowClient = field(default_factory=MlflowClient)
    log_system_metrics: bool = True
    silent: bool = False
    create_runs: bool = True
    extra_tags: Optional[Dict[str, str]] = None
    getml_project_path: Optional[Path] = None


@dataclass
class LoggingConfiguration:
    general: GeneralLoggingConfiguration = field(
        default_factory=GeneralLoggingConfiguration
    )
    data_container: DataContainerLoggingConfiguration = field(
        default_factory=DataContainerLoggingConfiguration
    )
    function: FunctionLoggingConfiguration = field(
        default_factory=FunctionLoggingConfiguration
    )
    pipeline: PipelineLoggingConfiguration = field(
        default_factory=PipelineLoggingConfiguration
    )



from __future__ import annotations

from typing import List


from getml_mlflow.logging import (
    datacontainer,
    function,
    logger,
    numpy,
    pipeline,
    systemmetrics,
)

__all__: List[str] = [
    "datacontainer",
    "function",
    "logger",
    "numpy",
    "pipeline",
    "systemmetrics",
]



from __future__ import annotations

import logging
from datetime import datetime, timezone
from logging import Logger
from types import TracebackType
from typing import Optional, Type

from mlflow import MlflowClient

logger: Logger = logging.getLogger(__name__)


def log_exit_exception(
    mlflow_client: MlflowClient,
    run_id: Optional[str],
    exc_type: Type[BaseException],
    exc_val: BaseException,
    exc_tb: Optional[TracebackType],
) -> None:
    if run_id is not None:
        mlflow_client.log_text(
            run_id=run_id,
            text=f"Exception: {exc_type}: {exc_val}",
            artifact_file=f"error/{datetime.now(timezone.utc).isoformat()}.log",
        )
    logger.error(
        f"Exception: {exc_type}: {exc_val}",
        exc_info=(exc_type, exc_val, exc_tb),
    )


def log_request_exception(
    mlflow_client: MlflowClient,
    run_id: str,
    exception: BaseException,
    context: str,
) -> None:
    mlflow_client.log_text(
        run_id=run_id,
        text=f"Exception: {context}: {exception}",
        artifact_file=f"error/{datetime.now(timezone.utc).isoformat()}.log",
    )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import getml
import mlflow
from mlflow import MlflowClient
from mlflow.utils.autologging_utils import autologging_integration
from mlflow.utils.autologging_utils.safety import revert_patches, safe_patch

from getml_mlflow.constants import DEFAULT_MLFLOW_TRACKING_URI
from getml_mlflow.flavor import FLAVOR_NAME
from getml_mlflow.loggingconfiguration import (
    DataContainerLoggingConfiguration,
    FunctionLoggingConfiguration,
    GeneralLoggingConfiguration,
    LoggingConfiguration,
    PipelineLoggingConfiguration,
)
from getml_mlflow.patch import engine, pipeline, project
from getml_mlflow.util.with_kwargs import with_kwargs


@dataclass
class SafePatchFunction:
    destination: Any
    function_name: str
    patch_function: Any
    with_logging_configuration: bool = True


FUNCTIONS_TO_PATCH: List[SafePatchFunction] = [
    SafePatchFunction(
        destination=getml,
        function_name="set_project",
        patch_function=engine.set_project,
    ),
    SafePatchFunction(
        destination=getml.engine,
        function_name="set_project",
        patch_function=engine.set_project,
    ),
    SafePatchFunction(
        destination=getml.engine.helpers,
        function_name="set_project",
        patch_function=engine.set_project,
    ),
    SafePatchFunction(
        destination=getml.project.attrs,
        function_name="switch",
        patch_function=project.switch,
    ),
    SafePatchFunction(
        destination=getml.pipeline.Pipeline,
        function_name="__init__",
        patch_function=pipeline.init,
        with_logging_configuration=False,
    ),
    SafePatchFunction(
        destination=getml.pipeline,
        function_name="load",
        patch_function=pipeline.load,
    ),
    SafePatchFunction(
        destination=getml.pipeline.helpers2,
        function_name="load",
        patch_function=pipeline.load,
    ),
    SafePatchFunction(
        destination=getml.pipeline.Pipeline,
        function_name="fit",
        patch_function=pipeline.fit,
    ),
    SafePatchFunction(
        destination=getml.pipeline.Pipeline,
        function_name="score",
        patch_function=pipeline.score,
    ),
    SafePatchFunction(
        destination=getml.pipeline.Pipeline,
        function_name="predict",
        patch_function=pipeline.predict,
    ),
    SafePatchFunction(
        destination=getml.pipeline.Pipeline,
        function_name="transform",
        patch_function=pipeline.transform,
    ),
]


@autologging_integration(FLAVOR_NAME)
def autolog(
    *,
    log_data_information: bool = True,
    log_data_as_artifact: bool = True,
    log_function_parameters: bool = True,
    log_function_return: bool = True,
    log_function_as_trace: bool = True,
    log_pipeline_parameters: bool = True,
    log_pipeline_tags: bool = True,
    log_pipeline_scores: bool = True,
    log_pipeline_features: bool = True,
    log_pipeline_columns: bool = True,
    log_pipeline_targets: bool = True,
    log_pipeline_data_model: bool = True,
    log_pipeline_as_artifact: bool = True,
    log_system_metrics: bool = True,
    disable: bool = False,
    silent: bool = False,
    create_runs: bool = True,
    extra_tags: Optional[Dict[str, str]] = None,
    getml_project_path: Optional[Path] = None,
    tracking_uri: Optional[str] = None,
) -> None:
    """Enable automatic logging of getML operations to MLflow.

    This function enables automatic logging of the following operations to MLflow:

    - pipeline creation, loading and operations (fit, score, predict, transform)
    - project setting and switching.

    Pipeline parameters, performance metrics, dataframe metadata, and other relevant
    information are captured and displayed in the MLflow UI. It also allows logging of
    dataframes passed as a function parameter or returned by a function as artifacts.

    The artifacts are stored in `artifacts-destination` set when running
    `mlflow ui` command. The default is `artifacts` directory in the current
    working directory.

    In the UI, getML pipelines correspond to MLflow runs, functions correspond to
    sub-runs, and projects correspond to experiments.

    For a detailed introduction on this MLflow integration, including setup, working with
    artifact pipelines, and more, please refer to our
    [Tracking with MLflow][mlflow-integration-guide] guide. The guide provides examples
    and configuration options to help you get the most from it.

    Args:
        log_data_information (bool, optional): Whether to log metadata about
            a `Dataframe` or `View` (e.g., number of rows & columns, column names, roles).

            The [`roles`][getml.data.roles] are indicated with the
            following emojis in the MLflow UI:

            - 🗃 for categorical columns
            - 🔗 for join keys
            - 🔢 for numerical columns
            - 🎯 for target column(s)
            - 📝 for text columns
            - ⏰ for timestamp columns
            - 🧮 for unused float columns
            - 🧵 for unused string columns

        log_data_as_artifact (bool, optional): Whether to log a `DataFrame`,
            `View` or `Subset` function parameter as a `.parquet` artifact. In addition,
            it allows logging of `DataFrame` returned by functions. In MLflow UI, the
            artifacts are available to download. `log_function_parameters` or
            `log_function_return` must be `True` for this to work.

        log_function_parameters (bool, optional): Whether to log parameters passed to
            getML functions, e.g., pipe.fit() in the MLflow UI. To log the `DataFrame`,
            `View` or `Subset` function parameters as artifacts,
            `log_data_as_artifact=True` must also be set to `True`.

        log_function_return (bool, optional): Whether to log return values of getML
            functions as artifacts. For example, it enables logging of `DataFrame`
            (as `.parquet`) and `numpy.ndarray` (as `.npy`) returned by `transform()`
            or `predict()` methods. `log_data_as_artifact` must also be `True`
            for `DataFrame` logging.

        log_function_as_trace (bool, optional): Whether to log function calls as MLflow
            traces for detailed execution flow.

        log_pipeline_parameters (bool, optional): Whether to log
            [`parameters`][getml.pipeline.Pipeline] of a pipeline.

        log_pipeline_tags (bool, optional): Whether to log
            [`tags`][getml.pipeline.Pipeline] of a pipeline.

        log_pipeline_scores (bool, optional): Whether to log [`scores`][getml.pipeline.Scores]
            (metrics) of a pipeline.

        log_pipeline_features (bool, optional): Whether to log [`features`][getml.pipeline.Features]
            learned during pipeline fitting.

        log_pipeline_columns (bool, optional): Whether to log [`columns`][getml.pipeline.Columns]
            (whose importance can be calculated) of a pipeline.

        log_pipeline_targets (bool, optional): Whether to log
            [`targets`][getml.pipeline.Pipeline.targets] of a pipeline.

        log_pipeline_data_model (bool, optional): Whether to log the
            [`data model`][getml.data.DataModel] provided in the pipeline. It is available
            as an HTML artifact to view or download.

        log_pipeline_as_artifact (bool, optional): Whether to save pipelines as
            MLflow artifacts.

            ??? note "Docker configuration, `download_artifact_pipeline()` and `switch_to_artifact_pipeline()`"

                When using this parameter with Docker, you'll need to set up proper bind
                mounts to allow pipeline artifact logging. For detailed instructions on
                working with artifact pipelines, Docker configurations, and related
                functions like `download_artifact_pipeline()` and `switch_to_artifact_pipeline()`,
                please refer to the [Tracking with MLflow][mlflow-integration-guide] guide.

        log_system_metrics (bool, optional): Whether to log system metrics (CPU, memory usage)
            during pipeline fitting. Metrics are available for getML Enterprise only.

        disable (bool, optional): If True, disables all getML autologging.

        silent (bool, optional): If True, suppresses all informational logging messages.

        create_runs (bool, optional): If True, creates new MLflow runs automatically
            when logging. You may set it to False and log under your own run. For example:

            ```python
            import mlflow
            mlflow.set_tracking_uri("http://localhost:5000")
            mlflow.set_experiment("your_experiment_name")
            with mlflow.start_run(run_name="your_run_name"):
                pipe.fit(container.train)
            ```

        extra_tags (Dict[str, str], optional): Additional custom tags to log with each MLflow run.

        getml_project_path (Path, optional): Path to the getML projects directory.
            Used for accessing and logging pipeline artifacts when
            `log_pipeline_as_artifact=True`. If not provided, defaults to
            `$HOME/.getML/projects`.

        tracking_uri (str, optional): MLflow tracking server URI. If not provided,
            uses `http://localhost:5000`.


    Examples:
        Basic usage with default settings:

        ```python
        import getml
        import getml_mlflow
        getml_mlflow.autolog()
        # Subsequent getML pipeline operations will be logged to MLflow
        ```

        Custom configuration:
        ```python
        getml_mlflow.autolog(
            log_pipeline_as_artifact=True,
            log_system_metrics=False,
            tracking_uri="http://localhost:5000"
        )
        ```
    """
    if disable:
        revert_patches(FLAVOR_NAME)
        return

    tracking_uri = tracking_uri or DEFAULT_MLFLOW_TRACKING_URI
    mlflow.set_tracking_uri(tracking_uri)

    logging_configuration: LoggingConfiguration = LoggingConfiguration(
        general=GeneralLoggingConfiguration(
            mlflow_client=MlflowClient(tracking_uri=tracking_uri),
            log_system_metrics=log_system_metrics,
            silent=silent,
            create_runs=create_runs,
            extra_tags=extra_tags,
            getml_project_path=getml_project_path,
        ),
        data_container=DataContainerLoggingConfiguration(
            log_information=log_data_information,
            log_as_artifact=log_data_as_artifact,
        ),
        function=FunctionLoggingConfiguration(
            log_parameters=log_function_parameters,
            log_return=log_function_return,
            log_as_trace=log_function_as_trace,
        ),
        pipeline=PipelineLoggingConfiguration(
            log_parameters=log_pipeline_parameters,
            log_tags=log_pipeline_tags,
            log_scores=log_pipeline_scores,
            log_features=log_pipeline_features,
            log_columns=log_pipeline_columns,
            log_targets=log_pipeline_targets,
            log_data_model=log_pipeline_data_model,
            log_as_artifact=log_pipeline_as_artifact,
        ),
    )

    for function in FUNCTIONS_TO_PATCH:
        safe_patch(
            autologging_integration=FLAVOR_NAME,
            destination=function.destination,
            function_name=function.function_name,
            patch_function=(
                with_kwargs(logging_configuration=logging_configuration)(
                    function.patch_function
                )
                if function.with_logging_configuration
                else function.patch_function
            ),
            manage_run=False,
        )



from __future__ import annotations

import logging
from collections import namedtuple
from logging import Logger
from threading import Event, Thread
from types import TracebackType
from typing import Callable, List, Optional, Type

import mlflow.utils.time
import numpy
import requests
from getml.constants import ENTERPRISE_DOCS_URL
from mlflow import MlflowClient
from mlflow.entities import Metric
from requests import Response
from requests.exceptions import RequestException

from getml_mlflow.constants import DEFAULT_GETML_HOST, DEFAULT_GETML_PORT
from getml_mlflow.logging.logger import log_exit_exception, log_request_exception
from getml_mlflow.util.callableenum import CallableEnumFactory

logger: Logger = logging.getLogger(__name__)


EngineMetric = CallableEnumFactory[Callable[[str], str]].build(
    "EngineMetric",
    dict(
        ENGINE_CPU_USAGE_PER_VIRTUAL_CORE_IN_PCT=lambda url: f"{url}/getcpuusage/",
        MEMORY_USAGE_IN_PCT=lambda url: f"{url}/getmemoryusage/",
    ),
)
ValidEngineMetric = namedtuple("ValidEngineMetric", ["name", "url"])


class SystemMetricsLogger:
    HOST: str = DEFAULT_GETML_HOST
    PORT: int = DEFAULT_GETML_PORT

    def __init__(
        self,
        mlflow_client: MlflowClient,
        run_id: str,
        host: str = HOST,
        port: int = PORT,
        *,
        log_system_metrics: bool = True,
    ) -> None:
        self._run_id: str = run_id
        self._event: Event = Event()
        self._thread: Optional[Thread] = None

        self._url: str = f"http://{host}:{port}"
        self._mlflow_client: MlflowClient = mlflow_client
        self._log_system_metrics: bool = log_system_metrics

    def _run_logging_metrics(self) -> None:
        step: int = 0
        valid_engine_metrics: List[ValidEngineMetric] = self._valid_metrics()
        while not self._event.is_set():
            self._log_metrics(step, valid_engine_metrics)
            step += 1
            self._event.wait(1)

    def _log_metrics(
        self, step: int, valid_engine_metrics: List[ValidEngineMetric]
    ) -> None:
        metrics: List[Metric] = []
        timestamp: int = mlflow.utils.time.get_current_time_millis()
        for engine_metric_name, engine_metric_url in valid_engine_metrics:
            try:
                response: Response = requests.get(engine_metric_url)
                metrics.append(
                    Metric(
                        key=engine_metric_name,
                        value=numpy.round(response.json()["data"][0][-1], 2),
                        timestamp=timestamp,
                        step=step,
                    )
                )
            except RequestException as exception:
                log_request_exception(
                    self._mlflow_client,
                    self._run_id,
                    exception,
                    f"GET({engine_metric_url})",
                )
                continue
        if metrics:
            self._mlflow_client.log_batch(
                run_id=self._run_id,
                metrics=metrics,
            )

    def _valid_metrics(self) -> List[ValidEngineMetric]:
        valid_engine_metrics: List[ValidEngineMetric] = []
        for engine_metric in EngineMetric:
            engine_metric_endpoint: str = engine_metric.value(self._url)
            try:
                response: Response = requests.get(engine_metric_endpoint)
                if response.ok:
                    valid_engine_metrics.append(
                        ValidEngineMetric(
                            engine_metric.name.lower(), engine_metric_endpoint
                        )
                    )
                else:
                    response.raise_for_status()
            except requests.exceptions.RequestException as exception:
                log_request_exception(
                    self._mlflow_client,
                    self._run_id,
                    exception,
                    f"GET({engine_metric_endpoint})",
                )
                logger.warning(
                    f"Engine metrics ({engine_metric_endpoint}) are available in the Enterprise edition. "
                    f"Visit {ENTERPRISE_DOCS_URL} for more information"
                )
                continue
        return valid_engine_metrics

    def __enter__(self) -> SystemMetricsLogger:
        if not self._log_system_metrics:
            return self

        self._thread = Thread(target=self._run_logging_metrics)
        self._thread.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        if exc_type is not None and exc_value is not None:
            log_exit_exception(
                self._mlflow_client, self._run_id, exc_type, exc_value, exc_traceback
            )

        if not self._log_system_metrics:
            return

        self._event.set()
        if self._thread is not None:
            self._thread.join()
        self._thread = None



from __future__ import annotations

from tempfile import TemporaryDirectory
from typing import Optional

import numpy
from mlflow import MlflowClient
from numpy.typing import NDArray


class NumpyLogger:
    def __init__(
        self,
        mlflow_client: MlflowClient,
        run_id: str,
    ) -> None:
        self._mlflow_client: MlflowClient = mlflow_client
        self._run_id: str = run_id

    def log_ndarray_as_artifact(
        self,
        data: NDArray[numpy.float_],
        name: str,
        artifact_path: Optional[str] = None,
    ) -> None:
        with TemporaryDirectory() as temp_dir:
            filename: str = f"{name}.npy"
            local_path: str = f"{temp_dir}/{filename}"
            numpy.save(local_path, data)
            self._mlflow_client.log_artifact(
                run_id=self._run_id,
                local_path=local_path,
                artifact_path=artifact_path,
            )



from __future__ import annotations

import inspect
import logging
from functools import cached_property, wraps
from inspect import BoundArguments
from logging import Logger
from typing import Any, Callable, Dict, Optional, OrderedDict, Sequence, TypeVar

import numpy
from getml import Pipeline
from getml.data import DataFrame, Subset
from getml.pipeline import Scores
from mlflow import MlflowClient
from mlflow.entities import Param, Run, Span
from mlflow.tracing.constant import TraceMetadataKey
from mlflow.tracing.trace_manager import InMemoryTraceManager
from typing_extensions import ParamSpec

from getml_mlflow.data.dataframelike import DataFrameLikeT
from getml_mlflow.logging.datacontainer import DataContainerLogger
from getml_mlflow.logging.numpy import NumpyLogger
from getml_mlflow.loggingconfiguration import (
    DataContainerLoggingConfiguration,
    FunctionLoggingConfiguration,
)

logger: Logger = logging.getLogger(__name__)

CallableArgsTypes = ParamSpec("CallableArgsTypes")
CallableReturnType = TypeVar("CallableReturnType")


class FunctionLogger:
    def __init__(
        self,
        mlflow_client: MlflowClient,
        run: Run,
        pipeline: Pipeline,
        *,
        function_logging_configuration: FunctionLoggingConfiguration = FunctionLoggingConfiguration(),
        data_container_logging_configuration: DataContainerLoggingConfiguration = DataContainerLoggingConfiguration(),
    ) -> None:
        self._mlflow_client: MlflowClient = mlflow_client
        self._run: Run = run
        self._pipeline: Pipeline = pipeline
        self._function_logging_configuration: FunctionLoggingConfiguration = (
            function_logging_configuration
        )
        self._data_container_logging_configuration: DataContainerLoggingConfiguration = data_container_logging_configuration

    def log(
        self, function: Callable[CallableArgsTypes, CallableReturnType]
    ) -> Callable[CallableArgsTypes, CallableReturnType]:
        @wraps(function)
        def wrapper(
            *args: CallableArgsTypes.args, **kwargs: CallableArgsTypes.kwargs
        ) -> CallableReturnType:
            bound_arguments: BoundArguments = inspect.signature(function).bind(
                *args, **kwargs
            )
            bound_arguments.apply_defaults()
            arguments: OrderedDict[str, Any] = bound_arguments.arguments

            span: Optional[Span] = self.log_trace_start(function.__name__, arguments)
            self.log_parameters(arguments)

            output: CallableReturnType = function(*args, **kwargs)

            self.log_return(output)
            self.log_trace_end(span, output)

            return output

        return wrapper

    def log_parameters(self, arguments: OrderedDict[str, Any]) -> None:
        if not self._function_logging_configuration.log_parameters:
            return

        parameters: Dict[str, Any] = {}

        for argument_name, argument_value in arguments.items():
            if argument_name == "self":
                # Don't log self.
                pass
            elif isinstance(argument_value, (DataFrameLikeT, Subset)):
                self._input_data_container_logger.log_data_container(
                    argument_value, [argument_name]
                )
            elif (
                isinstance(argument_value, Sequence)
                and all(
                    isinstance(element, DataFrameLikeT) for element in argument_value
                )
            ) or (
                isinstance(argument_value, Dict)
                and all(
                    isinstance(element, DataFrameLikeT)
                    for element in argument_value.values()
                )
            ):
                self._input_data_container_logger.log_data_containers(
                    argument_value, argument_name
                )
            else:
                parameters[argument_name] = str(argument_value)

        self._mlflow_client.log_batch(
            self._run.info.run_id,
            params=[Param(key, value) for key, value in parameters.items()],
        )

    @cached_property
    def _input_data_container_logger(self) -> DataContainerLogger:
        return DataContainerLogger.as_input(
            self._mlflow_client,
            self._run.info.run_id,
            logging_configuration=self._data_container_logging_configuration,
        )

    @cached_property
    def _artifact_data_container_logger(self) -> DataContainerLogger:
        return DataContainerLogger.as_artifact(
            self._mlflow_client,
            self._run.info.run_id,
            logging_configuration=self._data_container_logging_configuration,
        )

    @cached_property
    def _numpy_logger(self) -> NumpyLogger:
        return NumpyLogger(self._mlflow_client, self._run.info.run_id)

    def log_return(self, output: Any) -> None:
        if not self._function_logging_configuration.log_return or output is None:
            return

        if isinstance(output, DataFrame):
            self._artifact_data_container_logger.log_data_container(
                data_container=output,
                context="output",
            )
        elif isinstance(output, numpy.ndarray):
            self._numpy_logger.log_ndarray_as_artifact(
                data=output,
                name="output",
                artifact_path="output",
            )
        elif isinstance(output, (Pipeline, Scores)):
            # Return values of type Pipeline and Scores are not logged.
            pass
        else:
            logger.info("Missing return logging for type '%s'", type(output))

    def log_trace_start(
        self, function_name: str, arguments: OrderedDict[str, Any]
    ) -> Optional[Span]:
        if not self._function_logging_configuration.log_as_trace:
            return None

        span: Span = self._mlflow_client.start_trace(
            function_name,
            inputs=arguments,
            experiment_id=self._run.info.experiment_id,
            attributes={
                "pipeline": self._pipeline.id,
                "run": self._run.info.run_id,
            },
            tags={
                "pipeline": self._pipeline.id,
                "run": self._run.info.run_id,
            },
        )
        InMemoryTraceManager.get_instance().set_request_metadata(
            span.request_id,
            TraceMetadataKey.SOURCE_RUN,
            self._run.info.run_id,
        )
        return span

    def log_trace_end(self, span: Optional[Span], output: Any) -> None:
        if not self._function_logging_configuration.log_as_trace or span is None:
            return

        self._mlflow_client.end_trace(
            span.request_id,
            outputs={output.__class__.__name__: output},
            attributes={
                "pipeline": self._pipeline.id,
            },
        )
        self._mlflow_client.set_trace_tag(
            span.request_id, "pipeline", self._pipeline.id
        )



from __future__ import annotations

from enum import Enum, auto
from typing import Any, Callable, Dict, Generic, Type, TypeVar, cast

CallableType = TypeVar("CallableType", bound=Callable[..., Any])


class CallableEnumFactory(Generic[CallableType]):
    @classmethod
    def build(cls, name: str, entries: Dict[str, CallableType]) -> Type[Enum]:
        enum = Enum(name, {key: (auto(), value) for key, value in entries.items()})
        setattr(enum, "value", property(lambda self: self._value_[1]))
        return cast(Type[Enum], enum)



from __future__ import annotations

import functools
from typing import Any, Callable, TypeVar

from typing_extensions import ParamSpec

CallableArgsTypes = ParamSpec("CallableArgsTypes")
CallableReturnType = TypeVar("CallableReturnType")


def with_kwargs(
    **extra_kwargs: Any,
) -> Callable[
    [Callable[CallableArgsTypes, CallableReturnType]],
    Callable[CallableArgsTypes, CallableReturnType],
]:
    def decorator(
        function: Callable[CallableArgsTypes, CallableReturnType],
    ) -> Callable[CallableArgsTypes, CallableReturnType]:
        @functools.wraps(function)
        def wrapper(
            *args: CallableArgsTypes.args, **kwargs: CallableArgsTypes.kwargs
        ) -> CallableReturnType:
            kwargs.update(extra_kwargs)
            return function(*args, **kwargs)

        return wrapper

    return decorator
